@@ -9,6 +9,7 @@
 #include "lib/conf.h"
 #include "lib/lists.h"
 #include "lib/mempool.h"
+#include "lib/hashfunc.h"
 #include "sherlock/object.h"
 #include "gather/daemon/gatherd.h"
 
@@ -61,14 +62,19 @@ process_refs(struct odes *o)
 
 /*** Dynamic filtering by robots.txt ***/
 
+static struct buck2obj_buf *read_buf;
+  /* Used at three places, but never at the same time:
+   * 1. the parent process parses the replies of its children here
+   * 2. the child process first parses the robots.txt file here
+   * 3. and, then, it reads the original object info gthis->refreshing via this buffer
+   */
+
 static int
-dynamic_filter(byte *urlrest, oid_t oid)
+dynamic_filter(struct mempool *pool, byte *urlrest, oid_t oid)
 {
-  struct fastbuf *b;
-  byte line[1024], rest[MAX_URL_SIZE];
+  byte rest[MAX_URL_SIZE];
   struct obuck_header buck;
   int accept = 1;
-  byte *end;
 
   if (oid == OID_UNDEFINED || oid >= OID_FIRST_ERROR)
     return 1;
@@ -76,10 +82,10 @@ dynamic_filter(byte *urlrest, oid_t oid)
     die("dynamic_filter: error escaping %s", urlrest);
   buck.oid = oid;
   obuck_find_by_oid(&buck);
-  ASSERT(buck.type <= BUCKET_TYPE_PLAIN);
-  b = obuck_fetch();
-  while ((end = bgets(b, line, sizeof(line))) && accept)
-    if (line[0] == 'r' && !strncmp(rest, line+1, end-line-1))
+  struct fastbuf *b = obuck_fetch();
+  struct odes *obj = obj_read_bucket(read_buf, pool, buck.type, buck.length, b, NULL);
+  for (struct oattr *o = obj_find_attr(obj, 'r'); o; o=o->same)
+    if (!strncmp(rest, o->val, str_len(o->val)))
       {
 	accept = 0;
 	break;
@@ -193,22 +199,20 @@ gather_document(struct thread *t)
   gthis->error_hook = error_hook;
   gthis->url = gobj_parse_url(&gthis->url_s, t->url, "document", 0);
   gthis->robot_file_p = t->fetching_robot_file;
+  if (!t->fetching_robot_file && !dynamic_filter(t->pool, gthis->url_s.rest, t->host->robot_id))
+    gerror(2304, "Forbidden by robots.txt");
   if (t->refreshing != OID_UNDEFINED)
     {
       /* Read the old object we're trying to refresh */
       struct obuck_header oh;
-      struct fastbuf *b;
       oh.oid = t->refreshing;
       obuck_find_by_oid(&oh);
-      b = obuck_fetch();
-      gthis->refreshing = obj_new(gthis->pool);
-      obj_read(b, gthis->refreshing);
+      struct fastbuf *b = obuck_fetch();
+      gthis->refreshing = obj_read_bucket(read_buf, gthis->pool, oh.type, oh.length, b, NULL);
       bclose(b);
       if (t->ur.http_last_mod)
 	obj_set_attr_num(gthis->refreshing, 'L', t->ur.http_last_mod);
     }
-  if (!t->fetching_robot_file && !dynamic_filter(gthis->url_s.rest, t->host->robot_id))
-    gerror(2304, "Forbidden by robots.txt");
   if (!t->host->qkey)
     {
       setproctitle("R %.64s", t->url);
@@ -450,7 +454,6 @@ static volatile int shut_down;
 static struct pollfd *poll_table;
 static struct thread **poll_to_thread;
 static int poll_count;
-static struct buck2obj_buf *read_buf;
 
 static void				/* Just sets the "shut_down" variable */
 sig_int(int x UNUSED)
@@ -482,8 +485,6 @@ init_threads(void)
   sigaction(SIGTERM, &siga, NULL);
   siga.sa_handler = SIG_IGN;
   sigaction(SIGHUP, &siga, NULL);
-
-  read_buf = buck2obj_alloc();
 }
 
 static void
@@ -768,6 +769,7 @@ main(int argc, char **argv)
   refs_init(0);
   obuck_init(1);
 
+  read_buf = buck2obj_alloc();
   init_threads();
   gather_note_state("gathering");
   loop();
